/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CopyFilter.h"
#include "QDebug.h"
#include "../eMountWin.h"
#include "../eMountSys.h"
#include "../Util.h"
#include "../QConfig.h"

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <glibmm/i18n.h>
#include <glibmm/timeval.h>

#define COPYBUFLEN 		(unsigned int)131072 //128KB

namespace Pline 
{

//==============================================================================

CopyFilter::CopyFilter()
{
	gpoCloneData = 0;
	gpfCurrentSpeed = 0;
	gpiPctDone = 0;
	gpstETA = 0;
	
	gtStatSizeTotal = 0;
	gullSizeByte = 0;
	
	geRunMode = kNoRunMode;
	geRunModeLast = kNoRunMode;
}

//==============================================================================

CopyFilter::~CopyFilter()
{
}

//==============================================================================

void CopyFilter::askData()
{
	Glib::ustring stReason;
	
	if (eMountSys::getInstance().getRunMode() != eMountSys::kRunGui)
		return Filter::askData();
	 
	if (!isValidData(&stReason))
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
		
		tConfigItem.bAutoRun = false;
		tConfigItem.bWaitData = false;
		tConfigItem.bAskOverwrite = true;
		tConfigItem.stOverwriteData = gpoCloneData->getDest();
		tConfigItem.bOverwrite = gpoCloneData->getCanOverwrite();
		tConfigItem.stMsg = stReason;
		tConfigItem.stLabel = gpoCloneData->getLabel();
		tConfigItem.stSrc = gpoCloneData->getSrc();
		tConfigItem.stCommand = getFilterName();
		
		VolumeDialog oDlg(tConfigItem);
		
		if (oDlg.run() != Gtk::RESPONSE_OK)
		{
			oDlg.hide();
		
			while (Gtk::Main::events_pending()) // force redraw
				Gtk::Main::iteration();
			
			cancelDataReq();
			return;
		}

		gpoCloneData->setDest(oDlg.getOverwriteData());
		gpoCloneData->setCanOverwrite(oDlg.getOverwriteFlag());
					
		oDlg.hide();
		
		// save default destination
		QConfig::getInstance().setDfCloneDir(
			Glib::path_get_dirname(oDlg.getOverwriteData()));

		while (Gtk::Main::events_pending()) // force redraw
			Gtk::Main::iteration();
	}
	
	Filter::askData();
}

//==============================================================================

void CopyFilter::calculateETA()
{
	assert(gpfCurrentSpeed);
	assert(gpstETA);
	assert(gpiPctDone);
	
	if (*gpfCurrentSpeed == 0)
	{
		*gpstETA = _("inf");
		return;
	}
	
	off64_t tSizeCopyMB = gtStatSizeTotal / 1024 / 1024;
	unsigned long ulTotalSec = tSizeCopyMB / *gpfCurrentSpeed;
	unsigned long ulEtaSec = ulTotalSec - (ulTotalSec * ((float) *gpiPctDone / (float)100));
	
	// set time string
	time_t tTime = (time_t)ulEtaSec;
	struct tm * timeinfo;
	char szTimeStr[10]; memset(szTimeStr, '\0', 10);

	timeinfo = gmtime(&tTime);	
	int iHours = ((timeinfo->tm_mday - 1) * 24) + timeinfo->tm_hour;
	
	strftime(szTimeStr, 10, "%M:%S", timeinfo);

	Glib::ustring stETA = Util::compose("%d:%s", iHours, szTimeStr);
	
	*gpstETA = stETA;
}

//==============================================================================

bool CopyFilter::copy(	const Glib::ustring& stSrc,
						const Glib::ustring& stDest,
						const bool bEnlarge, 
						const bool bWithLimit)
{
	assert(gpoCloneData);
	assert(gpfCurrentSpeed);
	assert(gpiPctDone);
		
	bool bErr = false;
	
	int fdIn = 0;
	int fdOut = 0;
	
	unsigned char puyBuf[COPYBUFLEN];
	
	Glib::TimeVal oTimeStart;
	Glib::TimeVal oTimeCurr;
	unsigned long ulMilliSecDiff = 0;
	float fMBSec = 0;
	
	off64_t tSrcFileSize = 0;
	off64_t tRead;
	off64_t tSumWritten = 0;
	off64_t tOneSec = 0;
	
	setCanCancel(true);
	
	// reset stats
	gtStatSizeTotal = 0;
	
	// open source
	fdIn = open(stSrc.c_str(), O_RDONLY|O_LARGEFILE);
		
	if (fdIn < 0)
	{
		gstLastError = Util::compose(_("Could not open %s."), stSrc.c_str());
		
		bErr = true;
		goto EXIT;
	}

	// set tSrcFileSize
	if (bEnlarge || bWithLimit)
		tSrcFileSize = gullSizeByte;
	else
		tSrcFileSize = lseek(fdIn, 0, SEEK_END);
	
	gtStatSizeTotal = tSrcFileSize;
	
	if (tSrcFileSize <= 0)
	{
		gstLastError = Util::compose(_("%s: unknown size."), stSrc.c_str());
		
		bErr = true;
		goto EXIT;
	}
	
	// move to source's start
	lseek(fdIn, 0, SEEK_SET);

	// open output
	if (bEnlarge)
	{
		fdOut = open(stDest.c_str(), O_LARGEFILE|O_WRONLY, S_IRUSR|S_IWUSR);
		
		if (fdOut >= 0 && lseek(fdOut, 0, SEEK_END) < 0)
		{
			gstLastError = Util::compose(_("Can't read destinantion %s."), stDest.c_str());
			
			close(fdOut);
			
			bErr = true;
			goto EXIT;
		}
	}
	else
		fdOut = open(stDest.c_str(), O_CREAT|O_LARGEFILE|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR);
	
	if (fdOut < 0)
	{
		gstLastError = Util::compose(_("Can't read destinantion %s."), stDest.c_str());
		
		bErr = true;
		goto EXIT;
	}

	// copy
	oTimeStart.assign_current_time();
	
	update();
					
	while (true)
	{	
		off64_t tReadCount = COPYBUFLEN;

		if (gbStop)
			break;
									
		memset(puyBuf, 0, COPYBUFLEN);

		if (tSumWritten+tReadCount > tSrcFileSize)
			tReadCount = tSrcFileSize-tSumWritten;

		tRead = (tReadCount > 0) ? read(fdIn, puyBuf, tReadCount) : 0;
		
		if (tRead < 0)
		{
			gstLastError = Util::compose(_("Can't read source %s."), stSrc.c_str());
			
			bErr = true;
			break;
		}
		else if (tRead == 0 || tSumWritten >= tSrcFileSize)
			break;
		else
		{
			off64_t tWrite;
		
			tWrite = write(fdOut, puyBuf, tRead);
		
			tSumWritten += tWrite;
			tOneSec += tWrite;
			
			if (tRead != tWrite)
			{
				bErr = true;
				break;
			}

			*gpiPctDone = (int)((float)tSumWritten*(float)100/(float)tSrcFileSize);
			
			oTimeCurr.assign_current_time();
			ulMilliSecDiff = (unsigned long)((oTimeCurr - oTimeStart).as_double() * 1000);
			
			if (ulMilliSecDiff >= 1500)
			{
				fMBSec = (float)tOneSec/(float)1024/(float)1024/(float)ulMilliSecDiff*(float)1000;
				ulMilliSecDiff = 0;
				tOneSec = 0;
				
				oTimeStart.assign_current_time();
				
				*gpfCurrentSpeed = fMBSec;
				
				calculateETA();
				
				update();
			}							
		}
	}
	
	close(fdOut);

EXIT:
	close(fdIn);
	
	return (gbStop) ? false : !bErr;
}

//==============================================================================

bool CopyFilter::isValidData(Glib::ustring *pstReason)
{
	assert(gpoCloneData);
	
	if (pstReason)
		pstReason->clear();
		
	if (gpoCloneData->getDest().empty())
		return false;
	
	if (geRunModeLast != kRunCopyEnlargeExpand
		&& Util::isFileExist(gpoCloneData->getDest())
		&& !gpoCloneData->getCanOverwrite())
	{
		if (pstReason)
			*pstReason = _("Confirm Overwrite");
			
		return false;
	}
	
	return true;
}

//==============================================================================

bool CopyFilter::run()
{
	assert(gpoCloneData);
	
	assert(gpfCurrentSpeed);
	assert(gpiPctDone);
	
	setCanCancel(false);
	
	// reset error
	gstLastError.clear();
	
	// run
	geRunModeLast = geRunMode;
	
	// set run mode to no mode
	geRunMode = kNoRunMode;
		
	// ask destination
	while (true)
	{
		if (!isValidData())
		{
			requestData();	

			if (!waitData())
				return false;
		}
		else
			break;
	}

	const Glib::ustring stSrc = gpoCloneData->getSrc();
	const Glib::ustring stDest = gpoCloneData->getDest();
		
	if (geRunModeLast == kNoRunMode)
		return false;
	else if (geRunModeLast == kRunCopyFileStartNew)
	{
		return copy(	stSrc,
						stDest,
						false, 
						false);
	}
	else if (geRunModeLast == kRunCopyWithLimitStartNew)
	{
		if (gullSizeByte == 0)
			return false;
			
		return copy(	stSrc,
						stDest,
						false, 
						true);
	}
	else if (geRunModeLast == kRunTruncateWithLimitStartNew)
	{
		if (gullSizeByte == 0)
			return false;
		
		return truncate(stDest);
	}
	else if (geRunModeLast == kRunCopyEnlargeExpand)
	{
		if (gullSizeByte == 0)
			return false;
			
		return copy(	stSrc,
						stDest,
						true, 
						true);
	}
	
	return false;
}

//==============================================================================

void CopyFilter::setLimit(const unsigned long long ullVolumeSize)
{
	gullSizeByte = ullVolumeSize;
}

//==============================================================================

void CopyFilter::setSrcData(CloneData *poData)
{
	gpoCloneData = poData;
	
	Filter::setSrcData(poData);
}

//==============================================================================

bool CopyFilter::truncate(const Glib::ustring& stDest)
{
	int fdOut = open(stDest.c_str(), O_CREAT|O_LARGEFILE|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR);
	
	if (fdOut < 0)
	{
		gstLastError = Util::compose(_("Can't initialize destinantion %s."), stDest.c_str());	
		return false;
	}
	
	ssize_t twrite = write(fdOut, "\0", 1);
	
	close(fdOut);
	
	if (twrite >0 && truncate64(stDest.c_str(), gullSizeByte) == 0)
		return true;
	else
	{
		gstLastError = Util::compose(_("Can't create destinantion %s."), stDest.c_str());
		return false;
	}
}

//==============================================================================

} // Pline
