/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CryptUtil.h"
#include "QDebug.h"
#include "../Util.h"

#include <glibmm/i18n.h>
#include <glibmm/fileutils.h>
#include <list>
#include <sys/utsname.h>

namespace Pline
{

#define CIPHERMOD_PATH_FMT	"/lib/modules/%s/kernel/crypto"
#define MAPPEDDEVICE_PATH	"/dev/mapper"
#define CRYPTSETUP_PATH		"/sbin/cryptsetup"
#define CIPHERPROC_PATH 	"/proc/crypto"

//==============================================================================

CryptUtil::CryptUtil()
{
}

//==============================================================================

CryptUtil::~CryptUtil()
{
}

//==============================================================================

bool CryptUtil::detachDev(  const Glib::ustring& stDevMapper,
							Glib::ustring *pstErrMsg)
{
	Glib::ustring stCmd;
	Glib::ustring stOutBuf;
	
	if (pstErrMsg)
		pstErrMsg->clear();
	
	if (!Util::isFileExist(CRYPTSETUP_PATH, pstErrMsg))
		return false;

	if (!isValidDevice(stDevMapper, pstErrMsg))
		return false;
		
	stCmd = Util::compose(  "%s luksClose %s", 
							CRYPTSETUP_PATH,
							stDevMapper.c_str());
	
	// exec command
	if (Util::exec(stCmd, &stOutBuf))
	{
		debug_message(DEBUG_PLINE, "dev_mapper [%s]", stDevMapper.c_str());
				
		return true; //isLuks(stDevMapper);
	}
	else if (pstErrMsg)
	{
		size_t iPos = 0;
		
		while ((iPos = stOutBuf.find("\n")) != std::string::npos)
			stOutBuf.replace(iPos, 1, "");
			
		*pstErrMsg = stOutBuf;
	}
	
	return false;
}

//==============================================================================

bool CryptUtil::findDev(const Glib::ustring& stDevice,
						Glib::ustring *pstDevMapper)
{
	std::vector<tDevMapperStatus> vecDevMapper;
	
	getDevMapperList(&vecDevMapper);
	
	for (   std::vector<tDevMapperStatus>::size_type xi = 0;
			xi < vecDevMapper.size();
			++xi)
	{
		tDevMapperStatus *ptDevMapperStatusItem = &(vecDevMapper[xi]);
		
		if (ptDevMapperStatusItem->stDevice == stDevice)
		{
			if (pstDevMapper)
				*pstDevMapper = ptDevMapperStatusItem->stDevMapper;
				
			return true;
		}
	}
	
	return false;
}

//==============================================================================

bool CryptUtil::formatCrypt(const Glib::ustring& stDev, 
							const Glib::ustring& stPasswd,
							const Glib::ustring& stKeyFile,
							const Glib::ustring& stCipher,
							const Glib::ustring& stCBC,
                            const Glib::ustring& stKeySize,
                            Glib::ustring *pstErrMsg)
{
	Glib::ustring stCBCCipher;
	Glib::ustring stArgs;
	Glib::ustring stCmd;
	Glib::ustring stOutBuf;
	Glib::ustring stPasswdDec;
	
	if (!Util::isFileExist(CRYPTSETUP_PATH, pstErrMsg))
		return false;
		
	// validate
	if (!isValidDevice(stDev, pstErrMsg))
		return false;

	if ((stPasswd.empty() && stKeyFile.empty())
		|| (!stPasswd.empty() && !stKeyFile.empty()))
	{
		*pstErrMsg = _("Password and keyfile are invalid.");
		
		return false;
	}

	if (!stKeyFile.empty())
	{
		if (!isValidKey(stKeyFile, pstErrMsg))
			return false;
	}
	
	// set cipher (including CBC)
	if (!stCipher.empty())
	{
		stCBCCipher = (stCBC.empty()) ? 
			stCipher : Util::compose("%s-%s", stCipher.c_str(), stCBC.c_str());
	}
	
	stArgs = "luksFormat";
	
	if (!stCBCCipher.empty())
		stArgs = Util::compose(	"%s -c \"%s\"", 
								stArgs.c_str(), 
								stCBCCipher.c_str());
	
	if (!stKeySize.empty())
		stArgs = Util::compose(	"%s --key-size %s", 
								stArgs.c_str(), 
								stKeySize.c_str());

	stArgs = Util::compose(	"%s -q %s", 
							stArgs.c_str(), 
							stDev.c_str());

	if (!stKeyFile.empty())
		stArgs = Util::compose("%s \"%s\"", 
								stArgs.c_str(), 
								stKeyFile.c_str());
								
	// build command
	stCmd = Util::compose("%s %s", CRYPTSETUP_PATH, stArgs.c_str());
	
	// add password
	if (!stPasswd.empty())
	{
		Util::base64Dec(stPasswd, &stPasswdDec);
		
		stCmd = Util::compose( "echo \"%s\" | %s",
								stPasswdDec.c_str(),
								stCmd.c_str());
	}
	
	// exec command
	if (Util::exec(stCmd, &stOutBuf))			
		return true;
	else if (pstErrMsg)
	{
		size_t iPos = 0;
	
		if (!stOutBuf.empty())
		{
			while ((iPos = stOutBuf.find("\n")) != std::string::npos)
				stOutBuf.replace(iPos, 1, "");
		}
			
		*pstErrMsg = stOutBuf;	
	}
	
	return false;
}
        
//==============================================================================

bool CryptUtil::getAvailCipherMod(std::vector<Glib::ustring> *pvecMod)
{
	Glib::ustring stCipherModPath = getCipherModPath();
	
	if (stCipherModPath.empty())
		return false;
		
	Glib::Dir dir(stCipherModPath);
	
	std::vector<Glib::ustring> vecEntry(dir.begin(), dir.end());
	
	for (   std::vector<Glib::ustring>::size_type xi = 0;
			xi<vecEntry.size();
			++xi)
	{
		// buil full path
		Glib::ustring stMod = Glib::build_filename(stCipherModPath, vecEntry[xi]);
			
		if (stMod.size()>3 && stMod.substr(stMod.size() - 3) == ".ko")
			pvecMod->push_back(stMod);
	}
 
	return (pvecMod->size() > 0) ? true : false;
}

//==============================================================================

void CryptUtil::getCBCList(std::vector<Glib::ustring> *pvecCBCMode)
{
	pvecCBCMode->push_back("cbc-essiv:sha256");
	pvecCBCMode->push_back("cbc-benbi");
	pvecCBCMode->push_back("cbc-plain");
	pvecCBCMode->push_back("xts-essiv:sha256");
	pvecCBCMode->push_back("xts-benbi");
	pvecCBCMode->push_back("xts-plain");
	pvecCBCMode->push_back("cbc-null");
	
	return;
}

//==============================================================================

void CryptUtil::getKeySizeList(std::vector<Glib::ustring> *pvecKeySize)
{
	// KeySize
	for (int xi=3; xi<11; xi++)
		pvecKeySize->push_back(Util::compose("%d", (int)pow(2, xi)));
	
	return;
}
		
//==============================================================================

bool CryptUtil::getCipherList(std::vector<Glib::ustring> *pvecCipher)
{
	Glib::ustring stProc;
	std::vector<Glib::ustring> vecCipher;
	
	if (!Util::file2string(CIPHERPROC_PATH, &stProc))
		return false;
	
	Glib::ustring stRe = "(.+?)[\t ]*:[\t ]*(.+?)\\R";
	Glib::RefPtr<Glib::Regex> re = Glib::Regex::create(stRe);
	
	if (!re->match(stProc))
		return false;
	
	std::vector<Glib::ustring> vec = re->split(stProc);
	Glib::ustring stName;
	Glib::ustring stType;
		
	for (   std::vector<Glib::ustring>::size_type xi = 0;
			xi<vec.size();
			++xi)
	{
		bool bCapped = false;
		
		// capture
		if (vec[xi].lowercase() == "name"
			&& xi+1 < vec.size())
		{
			stName = vec[xi+1];
			bCapped = true;
		}
		else if (	vec[xi].lowercase() == "type"
					&& xi+1 < vec.size())
		{
			stType = vec[xi+1];
			bCapped = true;
		}
		
		if (!bCapped)
			continue;
			
		if (stType == "cipher")
		{
			// push stName uniquely
			if (!Util::findSt(stName, vecCipher))
				vecCipher.push_back(stName);
				
			// clear captures
			stName.clear();
			stType.clear();
		}
	}
	
	sort(vecCipher.begin(), vecCipher.end(), Util::cmpst);
	
	for (   std::vector<Glib::ustring>::size_type xi = 0;
			xi<vecCipher.size();
			++xi)
	{
		pvecCipher->push_back(vecCipher[xi]);
	}
	
	return (vecCipher.size() > 0);
}

//==============================================================================

const Glib::ustring CryptUtil::getCipherModPath()
{
	struct utsname tBuf;
	Glib::ustring stSysRel;
	
	if (uname(&tBuf) == 0)
	{
		stSysRel = tBuf.release;
	
		return Util::compose(CIPHERMOD_PATH_FMT, stSysRel.c_str());
	}
	
	return "";
}

//==============================================================================

bool CryptUtil::getDeviceStatus(const Glib::ustring stDevMapper,
								tDevMapperStatus *ptDevMapperStatus)
{
	Glib::ustring stCmd;
	Glib::ustring stOutBuf;
	
	if (!Util::isFileExist(CRYPTSETUP_PATH))
		return false;

	if (!isValidDevice(stDevMapper))
		return false;
		
	stCmd = Util::compose(  "%s status %s", 
							CRYPTSETUP_PATH,
							stDevMapper.c_str());
	
	if (Util::exec(stCmd, &stOutBuf))
	{
		Glib::RefPtr<Glib::Regex> re = 
			Glib::Regex::create(".*?(\\w+):[ ]*(.*?)\\R");
		
		if (re->match(stOutBuf))
		{
			std::vector<Glib::ustring> vec = re->split(stOutBuf);
			tDevMapperStatus tDevMapperStatusItem;
			
			for (   std::vector<Glib::ustring>::size_type xi = 0;
					xi<vec.size();
					++xi)
			{
				Glib::ustring stToken = vec[xi].lowercase();
				
				if (stToken == "cipher" && xi < vec.size()-1)
					tDevMapperStatusItem.stCipher = vec[xi+1];
				else if (stToken == "keysize" && xi < vec.size()-1)
					tDevMapperStatusItem.stKeySize = vec[xi+1];
				else if (stToken == "device" && xi < vec.size()-1)
					tDevMapperStatusItem.stDevice = vec[xi+1];
				else if (stToken == "offset" && xi < vec.size()-1)
					tDevMapperStatusItem.stOffest = vec[xi+1];
				else if (stToken == "size" && xi < vec.size()-1)
					tDevMapperStatusItem.stSize = vec[xi+1];
				else if (stToken == "mode" && xi < vec.size()-1)
					tDevMapperStatusItem.stMode = vec[xi+1];
			}
			
			if (ptDevMapperStatus)
				*ptDevMapperStatus = tDevMapperStatusItem;
				
			return true;
		}
	}
	
	return false;
}

//==============================================================================

void CryptUtil::getDevMapperList(std::vector<tDevMapperStatus> *pvec)
{
	if (!Util::isDirectory(MAPPEDDEVICE_PATH))
		return;
		
	Glib::Dir dir(MAPPEDDEVICE_PATH);
	
	std::vector<Glib::ustring> vecEntry(dir.begin(), dir.end());
	
	for (   std::vector<Glib::ustring>::size_type xi = 0;
			xi<vecEntry.size();
			++xi)
	{
		// buil full path
		Glib::ustring stDevMapper = Glib::build_filename(MAPPEDDEVICE_PATH, vecEntry[xi]);
		
		// get status
		tDevMapperStatus tDevMapperStatusItem;
		
		if (getDeviceStatus(stDevMapper, &tDevMapperStatusItem))
		{
			tDevMapperStatusItem.stDevMapper = stDevMapper;
			
			pvec->push_back(tDevMapperStatusItem);
		}
	}
 
	return;
}

//==============================================================================

const Glib::ustring CryptUtil::getDevMapperName(const Glib::ustring stDevice)
{
	Glib::ustring stDevMapper;
	size_t iPos = 1;
	
	stDevMapper = stDevice;
	
	if (stDevMapper.length() < 1)
		return "";
		
	while ((iPos = stDevMapper.find("/", iPos)) != std::string::npos)
		stDevMapper.replace(iPos, 1, ".");
	
	stDevMapper = Glib::build_filename(MAPPEDDEVICE_PATH, stDevMapper);
	
	return stDevMapper;
}

//==============================================================================

bool CryptUtil::getStatus(	const Glib::ustring& stDev,
							Glib::ustring *pstCipher,
							Glib::ustring *pstKeySize)
{	
	Glib::ustring stCmd;
	Glib::ustring stOutBuf;
	
	tDevMapperStatus tDevMapperStatusItem;
	
	if (!getDeviceStatus(stDev, &tDevMapperStatusItem))
		return false;
	
	*pstCipher = tDevMapperStatusItem.stCipher;
	*pstKeySize = tDevMapperStatusItem.stKeySize;
	
	return true;
}
							
//==============================================================================

bool CryptUtil::isValidDevice( const Glib::ustring& stDev, 
								Glib::ustring *pstErrMsg)
{
	if (!Util::isFileExist(stDev, pstErrMsg))
		return false;
	
	if (!Util::isDevice(stDev, pstErrMsg))
		return false;
	
	return true;
}

//==============================================================================

bool CryptUtil::isValidKey( const Glib::ustring& stKey, 
							Glib::ustring *pstErrMsg)
{
	if (!Util::isFileExist(stKey, pstErrMsg))
		return false;
	
	if (!Util::isRegularFile(stKey, pstErrMsg))
		return false;
	
	return true;
}

//==============================================================================

bool CryptUtil::setCrypt(	const Glib::ustring& stDevice,
							const Glib::ustring& stPasswd,
							const Glib::ustring& stKey,
							Glib::ustring *pstDevMapper,
							Glib::ustring *pstErrMsg)
{
	Glib::ustring stCmd;
	Glib::ustring stOutBuf;
	Glib::ustring stDevMapper;
	Glib::ustring stDevMapperName;
	Glib::ustring stPasswdDec;
	
	if (!Util::isFileExist(CRYPTSETUP_PATH, pstErrMsg))
		return false;

	// check data
	if (!isValidDevice(stDevice, pstErrMsg))
		return false;

	if (stPasswd.empty() && stKey.empty())
	{
		if (pstErrMsg)
			*pstErrMsg = _("Password and keyfile are invalid.");
			
		return false;
	}

	// build command
	stCmd = Util::compose("%s luksOpen", CRYPTSETUP_PATH);
	
	if (!stKey.empty())
	{
		if (!isValidKey(stKey, pstErrMsg))
			return false;
		
		stCmd = Util::compose( "%s -d \"%s\"", 
								stCmd.c_str(), 
								stKey.c_str());
	}
	else
	{
		Util::base64Dec(stPasswd, &stPasswdDec);
		
		stCmd = Util::compose( "echo \"%s\" | %s",
								stPasswdDec.c_str(),
								stCmd.c_str());
	}
	
	if (pstDevMapper)
		pstDevMapper->clear();
		
	if (pstErrMsg)
		pstErrMsg->clear();
		
	stDevMapper = getDevMapperName(stDevice);
	stDevMapperName = Glib::path_get_basename(stDevMapper);
	
	if (stDevMapper.empty())
	{
		if (pstErrMsg)
			*pstErrMsg = "Invalid device mapper";
			
		return false;
	}
		
	stCmd = Util::compose(  "%s %s %s", 
							stCmd.c_str(), 
							stDevice.c_str(),
							stDevMapperName.c_str());
	
	// exec command
	if (Util::exec(stCmd, &stOutBuf))
	{
		debug_message(DEBUG_PLINE, "dev [%s], dev_mapper [%s]", stDevice.c_str(), stDevMapper.c_str());
	
		if (pstDevMapper)
			*pstDevMapper = stDevMapper;
			
		return true;
	}
	else if (pstErrMsg)
	{
		size_t iPos = 0;
	
		if (!stOutBuf.empty())
		{
			while ((iPos = stOutBuf.find("\n")) != std::string::npos)
				stOutBuf.replace(iPos, 1, "");
		}
		else
			stOutBuf = _("Unrecognized encrypted device.");
			
		*pstErrMsg = stOutBuf;
		
	}
	
	return false;
}

//==============================================================================

} // Pline	
