/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "LoopUtil.h"
#include "../QDebug.h"
#include "../Util.h"

#include <glibmm/i18n.h>
#include <sys/ioctl.h>
#include <linux/loop.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <grp.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>

#define LOOPDEVICE_PATH		"/dev"

#define LOSETUP_MAJOR 		(int)7
#define LOSETUP_MINLOOP		(unsigned int)16
#define LOSETUP_MAXLOOP		(unsigned int)256
#define LOSETUP_GROUPNAME 	"disk"
#define LOSETUP_LOOPFORMAT 	"/dev/loop%d"

using namespace std;

namespace Pline
{

//==============================================================================

LoopUtil::LoopUtil()
{
}

//==============================================================================

LoopUtil::~LoopUtil()
{
}

//==============================================================================

void LoopUtil::buildDevLoopList(std::vector<Glib::ustring> *pvecLoop)
{
	pvecLoop->clear();
	
	for (unsigned int xui=LOSETUP_MINLOOP; xui<LOSETUP_MAXLOOP; ++xui)
		pvecLoop->push_back(Util::compose(LOSETUP_LOOPFORMAT, (int)xui));
}

//==============================================================================

bool LoopUtil::createDev(Glib::ustring *pstDev)
{
	unsigned int xui;
	vector<Glib::ustring> vecLoop;
	Glib::ustring stDev;
	
	buildDevLoopList(&vecLoop);
	
	pstDev->erase();
	
	mode_t tMode = S_IFBLK|S_IRUSR|S_IWUSR|S_IWGRP|S_IRGRP;
	struct stat statbuf;
	
	for (xui=0; xui<vecLoop.size(); ++xui)
	{
		stDev = vecLoop[xui];
		
		if (stat(stDev.c_str(), &statbuf) != 0)
		{
			if (errno == ENOENT)
			{
				if (mknod(stDev.c_str(),
					tMode, 
					makedev(LOSETUP_MAJOR, xui+LOSETUP_MINLOOP)) == 0)
				{
					if (chmod(stDev.c_str(), S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP) == 0)
					{
						struct group *ptGroup = getgrnam(LOSETUP_GROUPNAME);
						
						if (ptGroup)
						{
							if (chown(stDev.c_str(), 0, (gid_t)ptGroup->gr_gid) == 0)
							{
								debug_message(DEBUG_PLINE, "dev [%s]", stDev.c_str());
								
								*pstDev = stDev;
								return true;
							}
						}
					}
				}
			}
		}
	}
	
	return false;
}

//==============================================================================

bool LoopUtil::detachDev(   const Glib::ustring& stDev,
							Glib::ustring* pstErrMsg)
{
	int fd;

	if ((fd = open(stDev.c_str(), O_RDONLY)) < 0)
	{
		if (pstErrMsg)
			*pstErrMsg = _("Failed to open device.");
			
		return false;
	}
	
	if (ioctl(fd, LOOP_CLR_FD, 0) < 0)
	{
		if (pstErrMsg)
			*pstErrMsg = _("Failed detaching loop device.");

		close(fd);
		return false;
	}
	close (fd);
	
	debug_message(DEBUG_PLINE, "dev [%s]", stDev.c_str());
	
	return true;
}

//==============================================================================

bool LoopUtil::findDev(	const Glib::ustring& stSrc,
						Glib::ustring *pstDev)
{
	unsigned int xui;
	vector<Glib::ustring> vecLoop;
	Glib::ustring stDev;
	Glib::ustring stSelDev;
	struct stat tSrcStat;
	
	getDevLoopList(&vecLoop);
	
	if (pstDev)
		pstDev->erase();
	
	if (stSrc.empty())
		return false;
	
	int fd;
	struct stat statbuf;
	struct loop_info loopinfo;
	struct loop_info64 loopinfo64;

	if (stat(stSrc.c_str(), &tSrcStat) != 0)
		return false;
		
	for (xui=0; xui<vecLoop.size(); ++xui)
	{
		stDev = vecLoop[xui];

		if (stat(stDev.c_str(), &statbuf) == 0 && S_ISBLK(statbuf.st_mode))
		{
			fd = open(stDev.c_str(), O_RDONLY);
			
			if (fd >= 0)
			{
				if (ioctl(fd, LOOP_GET_STATUS64, &loopinfo64) == 0)
				{
					if (loopinfo64.lo_device == tSrcStat.st_dev 
					    && loopinfo64.lo_inode == tSrcStat.st_ino)
					{
						close(fd);
						
						if (pstDev)
							*pstDev = stDev;

						return true;
					}
				}
				else if (ioctl(fd, LOOP_GET_STATUS, &loopinfo) != 0)
				{
					if (loopinfo.lo_device == tSrcStat.st_dev 
					    && loopinfo.lo_inode == tSrcStat.st_ino)
					{
						close(fd);
						
						if (pstDev)
							*pstDev = stDev;
								
						return true;
					}
				}
				
				close (fd);
			}
		}
	}
	
	return false;
}
						
//==============================================================================

bool LoopUtil::findUnusedDev(Glib::ustring *pstDev)
{
	unsigned int xui;
	vector<Glib::ustring> vecLoop;
	Glib::ustring stDev;
	
	int fd;
	struct stat statbuf;
	struct loop_info loopinfo;
	struct loop_info loopinfo64;
	
	getDevLoopList(&vecLoop);
	
	if (pstDev)
		pstDev->erase();
	
	for (xui=0; xui<vecLoop.size(); ++xui)
	{
		stDev = vecLoop[xui];
		
		if (stat(stDev.c_str(), &statbuf) == 0 && S_ISBLK(statbuf.st_mode))
		{
			fd = open(stDev.c_str(), O_RDONLY);
			
			if (fd >= 0)
			{
				if (ioctl(fd, LOOP_GET_STATUS64, &loopinfo64) != 0)
				{
					if (errno == ENXIO)
					{
						close(fd);
						
						if (pstDev)
							*pstDev = stDev;
							
						return true;
					}
				}
				else if (ioctl(fd, LOOP_GET_STATUS, &loopinfo) != 0)
				{
					if (errno == ENXIO)
					{
						close(fd);
							
						if (pstDev)
							*pstDev = stDev;
								
						return true;
					}
				}
				
				close (fd);
			}
		}
	}
	
	return false;
}

//==============================================================================

void LoopUtil::getDevLoopList(std::vector<Glib::ustring> *pvecLoop)
{
	Glib::Dir dir(LOOPDEVICE_PATH);

	std::vector<Glib::ustring> vecEntry(dir.begin(), dir.end());
	
	for (   std::vector<Glib::ustring>::size_type xi = 0;
			xi<vecEntry.size();
			++xi)
	{
		// buil full path
		Glib::ustring stDevice = Glib::build_filename(LOOPDEVICE_PATH, vecEntry[xi]);
		
		if (Glib::Regex::match_simple("loop[0-9]+", stDevice))
			pvecLoop->push_back(stDevice);
	}
}

//==============================================================================

bool LoopUtil::isValidSrc(  const Glib::ustring& stSrc, 
							Glib::ustring* pstErrMsg)
{
	if (!Util::isFileExist(stSrc, pstErrMsg))
		return false;
	
	if (!Util::isRegularFile(stSrc, pstErrMsg))
		return false;
	
	return true;
}

//==============================================================================

bool LoopUtil::isValidDev(  const Glib::ustring& stDev, 
							Glib::ustring* pstErrMsg)
{
	if (!Util::isFileExist(stDev, pstErrMsg))
		return false;
	
	if (!Util::isDevice(stDev, pstErrMsg))
		return false;
	
	return true;
}

//==============================================================================

bool LoopUtil::setLoop(	const Glib::ustring& stDev, 
						const Glib::ustring& stSrc,
						bool *pbReadOnly,
						Glib::ustring* pstErrMsg)
{
	
	struct loop_info loopinfo;
	struct loop_info64 loopinfo64;
	int fd, ffd, i;
	int mode = (*pbReadOnly ? O_RDONLY : O_RDWR);
	
	if (!isValidSrc(stSrc, pstErrMsg))
		return false;
	if (!isValidDev(stDev, pstErrMsg))
		return false;

	if ((ffd = open(stSrc.c_str(), mode)) < 0)
	{
		if (!*pbReadOnly && errno == EROFS)
			ffd = open(stSrc.c_str(), mode = O_RDONLY);
		
		if (ffd < 0)
			return false;
	}
	
	if ((fd = open(stDev.c_str(), mode)) < 0)
	{
		close(ffd);
		return false;
	}
	*pbReadOnly = (mode == O_RDONLY);
	
	memset(&loopinfo, 0, sizeof(loopinfo));
	memset(&loopinfo64, 0, sizeof(loopinfo64));

	strncpy((char*)loopinfo.lo_name, stSrc.c_str(), LO_NAME_SIZE);
	strncpy((char*)loopinfo64.lo_file_name, stSrc.c_str(), LO_NAME_SIZE);

	loopinfo.lo_offset = 0;
	loopinfo64.lo_offset = 0;
	loopinfo.lo_encrypt_key_size = 0;
	loopinfo64.lo_encrypt_key_size = 0;

	if (ioctl(fd, LOOP_SET_FD, ffd) < 0)
	{
		close(fd);
		close(ffd);
		return false;
	}
	close (ffd);

	i = ioctl(fd, LOOP_SET_STATUS64, &loopinfo64);
	if (i)
	{
		i = ioctl(fd, LOOP_SET_STATUS, &loopinfo);
		memset(&loopinfo, 0, sizeof(loopinfo));
	}
	memset(&loopinfo64, 0, sizeof(loopinfo64));

	if (i)
	{
		ioctl (fd, LOOP_CLR_FD, 0);
		close (fd);
		return false;
	}
	
	close (fd);

	debug_message(DEBUG_PLINE, "src [%s], dev [%s]", stSrc.c_str(), stDev.c_str());

	return true;
}

//==============================================================================

} // Pline
