/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "VolumeGui.h"
#include "QDebug.h"
#include "eMountWin.h"
#include "DispatchData.h"

#include <glibmm/i18n.h>
#include <assert.h>

//==============================================================================

VolumeGui::VolumeGui(VolumeConfig *poVolumeConfig)
: goVolumeConfig(*poVolumeConfig)
{
	// init
	gpoThread = 0;
	gbLockThread = false;
	gbStartPlineLocked = false;
	
	// create volume gui state
	gpoVolumeGuiState = new VolumeGuiState;
	
	// required for sort by function
	gpoVolumeGuiState->setLabel(goVolumeConfig.getData().getLabel());
	gpoVolumeGuiState->setPos(goVolumeConfig.getData().getPos());
			
	// dispatcher
	goDispatcher.connect(sigc::mem_fun(*this, &VolumeGui::onDispatcher));
}

//==============================================================================

VolumeGui::~VolumeGui()
{	
	assert(gbLockThread);
				
	if (gpoVolumeGuiState)
		delete gpoVolumeGuiState;
		
	g_return_if_fail(!gpoThread);
}

//==============================================================================

bool VolumeGui::canStartPline()
{
	if (isRunning())
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
			
		tConfigItem.bAutoRun = true;
		tConfigItem.bWaitData = false;
		tConfigItem.stLabel = goVolumeConfig.getData().getLabel();
		tConfigItem.stMsg = _("Volume busy.");
		
		VolumeDialog oDlg(tConfigItem);
		
		return false;
	}
	
	return true;
}

//==============================================================================

DispatchData* VolumeGui::createDispatchData()
{
	DispatchData *poDispatchData = new DispatchData;
	
	poDispatchData->signal_join().connect(
		sigc::mem_fun(*this, &VolumeGui::onJoin));
	
	poDispatchData->signal_pline_askdata().connect(
		sigc::mem_fun(*this, &VolumeGui::onPlineAskData));
	
	poDispatchData->signal_show_dialog().connect(
		sigc::mem_fun(*this, &VolumeGui::onShowDialog));
	
	poDispatchData->signal_update_volume().connect(
		sigc::mem_fun(*this, &VolumeGui::onUpdateVolume));
			
	return poDispatchData;
}

//==============================================================================

bool VolumeGui::isRunning()
{
	if (!gmtxLockThread.trylock())
		return true;
	
	bool bRunning = (gpoThread || gbLockThread || gbStartPlineLocked) ? true : false;

	gmtxLockThread.unlock();

	return bRunning;
}

//==============================================================================

void VolumeGui::join()
{	
	debug_message(DEBUG_ALL, "Completing thread [%s]...", goVolumeConfig.getData().getLabel().c_str());

	gpoThread->join();
	gpoThread = NULL;
	
	debug_message(DEBUG_ALL, "Thread completed [%s].", goVolumeConfig.getData().getLabel().c_str());
}

//==============================================================================

bool VolumeGui::lockStartPline()
{
	if (!gmtxLockStartPline.trylock())
		return false;
	
	if (gbStartPlineLocked)
	{
		gmtxLockStartPline.unlock();
		return false;
	}
	
	gbStartPlineLocked  = true;
	
	gmtxLockStartPline.unlock();
	
	return true;
}

//==============================================================================

void VolumeGui::launchThread()
{
	assert (gbStartPlineLocked);
	
	Glib::signal_idle().connect(
		sigc::bind_return(sigc::mem_fun(*this, &VolumeGui::launchThreadIdle), false));
}		
		
//==============================================================================

void VolumeGui::launchThreadIdle()
{
	assert(!gpoThread);
	assert (gbStartPlineLocked);
	
	if (!gmtxLockThread.trylock() || gbLockThread)
	{
		debug_message(DEBUG_ALL, "Threads are locked, ignoring request. %s", goVolumeConfig.getData().getLabel().c_str());
		return;
	}

	gmtxLockThread.unlock();

	// Create a joinable thread.
	gpoThread = Glib::Thread::create(
		sigc::mem_fun(*this, &VolumeGui::threadFun), true);
	
	// Unlock startpline
	Glib::Mutex::Lock lock(gmtxLockStartPline);
	
	gbStartPlineLocked = false;
}

//==============================================================================

bool VolumeGui::lockThread(const bool bLock)
{	
	if (!gmtxLockThread.trylock())
	{
		assert(bLock);
		
		return false;
	}
			
	if (gpoThread)
	{
		assert(bLock);
		
		gmtxLockThread.unlock();
		return false;
	}

	if (gbLockThread && bLock)
	{
		gmtxLockThread.unlock();
		return false;
	}
		
	// set lock
	gbLockThread = bLock;

	gmtxLockThread.unlock();
	
	return true;
}

//==============================================================================

void VolumeGui::onDispatcher()
{	
	eMountWin::processQueueData();
}

//==============================================================================

void VolumeGui::onJoin()
{
	join();
}

//==============================================================================

void VolumeGui::onPlineAskData(Pline::Pline* poPline)
{
	poPline->askData();
}

//==============================================================================

void VolumeGui::onShowDialog(VolumeDialog::tConfig tConfigDialogItem)
{
	tConfigDialogItem.bAutoRun = true;
	tConfigDialogItem.bWaitData = false;
			
	VolumeDialog oDlg(tConfigDialogItem);
}

//==============================================================================

void VolumeGui::onUpdateVolume()
{
	assert(gpoVolumeGuiState);
	
	eMountWin::updateVolumeGuiState(gpoVolumeGuiState);
}

//==============================================================================
