/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "VolumeTreeView.h"
#include "QDebug.h"
#include "QConfig.h"
#include "VolumeCellRenderer.h"
#include "VolumeGuiState.h"

#include <assert.h>

#include <glibmm/i18n.h>
#include <list>

//==============================================================================

VolumeTreeView::ModelColumns::ModelColumns()
: Gtk::TreeModel::ColumnRecord()
{
	add(goColVolumeCellRenderer);
	add(goColIndex);
	add(goColHeight);
	add(goColPos);
	add(goColBtnCancel);
	add(goColBtnEject);
	add(goColBtnMount);
	add(goColDrawInfo);
	add(goColDrawLabel);
	add(goColDrawProgressFull);
	add(goColDrawProgressPctGauge);
	add(goColDrawProgressPctUsed);
}

//==============================================================================
// VolumeTreeView
//==============================================================================

VolumeTreeView::VolumeTreeView(Gtk::ScrolledWindow *poScrollWin)
: gpoScrollWin(poScrollWin),
  gpoVolumeCellRenderer(0)
{
	// set ListStore model
	grefTreeModel = Gtk::ListStore::create(goColumns);
	set_model(grefTreeModel);

	// set treeview style
	set_headers_visible(false);
	set_enable_search(false);
	
	// set selection mode
	Glib::RefPtr<Gtk::TreeSelection> refTreeSelection = get_selection();
	refTreeSelection->set_mode(Gtk::SELECTION_SINGLE);
    
    // add motion event
	add_events(Gdk::POINTER_MOTION_MASK);
	
	// init guistate
	gtGuiState.iActiveIndex = -1;
	
    // connect selection signal
	refTreeSelection->signal_changed().connect(
		sigc::mem_fun(*this, &VolumeTreeView::on_selection_changed));
	
	// start a 10 sec timer
	goSigTimeoutConn = Glib::signal_timeout().connect(
				sigc::mem_fun(*this, &VolumeTreeView::on_timeout), 10000); // 10 sec.
}

//==============================================================================

VolumeTreeView::~VolumeTreeView()
{
	goSigTimeoutConn.disconnect();

	if (gpoVolumeCellRenderer)
		delete gpoVolumeCellRenderer;
}

//==============================================================================

void VolumeTreeView::addVolume(	VolumeGuiState *poVolumeGuiState, 
								const bool bSelectVolume)
{
	if (!gpoVolumeCellRenderer)
	{
		// create volcellrenderer column at fist item in
		Gtk::TreeView::Column *poColumn = Gtk::manage(new Gtk::TreeView::Column(""));

		gpoVolumeCellRenderer = new VolumeCellRenderer;
		
		poColumn->pack_start(*manage(gpoVolumeCellRenderer), true);
		
		// set properties
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_index(), goColumns.goColIndex);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_height(), goColumns.goColHeight);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_pos(), goColumns.goColPos);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_btn_cancel(), goColumns.goColBtnCancel);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_btn_eject(), goColumns.goColBtnEject);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_btn_mount(), goColumns.goColBtnMount);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_draw_info(), goColumns.goColDrawInfo);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_draw_label(), goColumns.goColDrawLabel);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_draw_progress_full(), goColumns.goColDrawProgressFull);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_draw_progress_pct_gauge(), goColumns.goColDrawProgressPctGauge);
		poColumn->add_attribute(
			gpoVolumeCellRenderer->property_draw_progress_pct_used(), goColumns.goColDrawProgressPctUsed);
					
		// append cols	
		append_column("index", goColumns.goColIndex);
		append_column("height", goColumns.goColHeight);
		append_column("pos", goColumns.goColPos);
		append_column("btn_cancel", goColumns.goColBtnCancel);
		append_column("btn_eject", goColumns.goColBtnEject);
		append_column("btn_mount", goColumns.goColBtnMount);
		append_column("info", goColumns.goColDrawInfo);
		append_column("label", goColumns.goColDrawLabel);
		append_column("progress_full", goColumns.goColDrawProgressFull);
		append_column("progress_pct_gauge", goColumns.goColDrawProgressPctGauge);
		append_column("progress_pct_used", goColumns.goColDrawProgressPctUsed);
		
		// set visibility
		std::list<Gtk::TreeViewColumn*> lstColumns = get_columns();
		
		for (std::list<Gtk::TreeViewColumn*>::iterator it = lstColumns.begin(); 
				it != lstColumns.end();
				it++)
		{
			((Gtk::TreeViewColumn*)*it)->set_visible(false);
		}
 
		// append renderer col
		append_column(*manage(poColumn));
		
		// create sort model
		if (!grefSortedModel)
		{
			grefSortedModel = Gtk::TreeModelSort::create(grefTreeModel);
	
			grefSortedModel->signal_rows_reordered().connect(
				sigc::mem_fun(*this, &VolumeTreeView::on_rows_reordered));
		}
		
		// grab activate signal
		gpoVolumeCellRenderer->signal_activate().connect(
			sigc::mem_fun(*this, &VolumeTreeView::on_cell_activate));
	}
	
	// append new row
	Gtk::TreeModel::Row oRow = *(grefTreeModel->append());
	
	// set cell renderer
	oRow[goColumns.goColVolumeCellRenderer] = gpoVolumeCellRenderer;
	
	// set row data
	mapVolumeGuiState(oRow, poVolumeGuiState);
																	
	// set row index
	oRow[goColumns.goColIndex] = gpoVolumeCellRenderer->createNewRowIndex();
	
	// add row to tVolumeGuiState
	tVolumeGuiState tVolumeGuiStateItem;
	
	tVolumeGuiStateItem.iIndex = oRow[goColumns.goColIndex];
	tVolumeGuiStateItem.poVolumeGuiState = poVolumeGuiState;
	
	glstVolumeGuiState.push_back(tVolumeGuiStateItem);
	
	// select and scroll to row
	if (bSelectVolume)
	{
		Glib::RefPtr<Gtk::TreeSelection> refTreeSelection = get_selection();

		Gtk::TreeModel::iterator ite_sorted = 
			grefSortedModel->convert_child_iter_to_iter(oRow);
	
		Gtk::TreeModel::Path row_path_sorted = grefSortedModel->get_path(ite_sorted);
	
		refTreeSelection->select(row_path_sorted);
		scroll_to_row(row_path_sorted);
	}
}

//==============================================================================

bool VolumeTreeView::getRowByIndex( const int iIndex,
									Gtk::TreeModel::Row *poRow)
{
	typedef Gtk::TreeModel::Children type_children; //minimise code length.

	type_children children = grefTreeModel->children();
	
	for (   type_children::iterator ite = children.begin();
			ite != children.end(); 
			++ite)
	{
		Gtk::TreeModel::Row row_ite = *ite;
		
		if (row_ite[goColumns.goColIndex] == iIndex)
		{
			*poRow = row_ite;
		
			return true;
		}
	}
	
	return false;
}

//==============================================================================

VolumeGuiState* VolumeTreeView::getSelectedVolume()
{
	if (!grefSortedModel)
		return 0;
	
	set_model(grefSortedModel);
	
	Glib::RefPtr<Gtk::TreeSelection> refTreeSelection = get_selection();
	Gtk::TreeModel::iterator ite_sorted = refTreeSelection->get_selected();
		
	if (ite_sorted)
    {
		Gtk::TreeModel::iterator ite =
			grefSortedModel->convert_iter_to_child_iter(ite_sorted);
		
		if (ite)
			return getVolumeByIndex((*ite)[goColumns.goColIndex]);
    }
	
	return 0;
}

//==============================================================================

VolumeGuiState* VolumeTreeView::getVolumeByIndex(const int iIndex)
{
	for (   std::list<tVolumeGuiState>::iterator ite = glstVolumeGuiState.begin();
			ite != glstVolumeGuiState.end();
			++ite)
	{
		if (((tVolumeGuiState)*ite).iIndex == iIndex)
			return ((tVolumeGuiState)*ite).poVolumeGuiState;
	}
	
	return 0;
}

//==============================================================================

void VolumeTreeView::mapVolumeGuiState( Gtk::TreeModel::Row& oRow,
										VolumeGuiState *poVolumeGuiState)
{
	// set row data
	
	oRow[goColumns.goColVolumeCellRenderer] = gpoVolumeCellRenderer;
	
	oRow[goColumns.goColPos] = poVolumeGuiState->getPos();
	oRow[goColumns.goColBtnCancel] = poVolumeGuiState->getBtnCancel();
	oRow[goColumns.goColBtnEject] = poVolumeGuiState->getBtnEject();
	oRow[goColumns.goColBtnMount] = poVolumeGuiState->getBtnMount();
	oRow[goColumns.goColDrawInfo] = poVolumeGuiState->getInfo();
	oRow[goColumns.goColDrawLabel] = poVolumeGuiState->getLabel();
	oRow[goColumns.goColDrawProgressFull] = poVolumeGuiState->getISO();
	oRow[goColumns.goColDrawProgressPctGauge] = poVolumeGuiState->getProgressPct();
	oRow[goColumns.goColDrawProgressPctUsed] = poVolumeGuiState->getUsedSpacePct();
	
	
	// calc row height
	bool bDrawProgress = false;
	
	if (poVolumeGuiState->getISO()
		|| poVolumeGuiState->getProgressPct() >= 0
		|| poVolumeGuiState->getUsedSpacePct() >= 0)
	{
		bDrawProgress = true;
	}

	oRow[goColumns.goColHeight] = gpoVolumeCellRenderer->calcRowHeight( poVolumeGuiState->getInfo(),
																		poVolumeGuiState->getLabel(),
																		bDrawProgress);
}

//==============================================================================

bool VolumeTreeView::on_button_press_event(GdkEventButton* event)
{	
	if (!event)
		return Gtk::TreeView::on_button_press_event(event);
			
	if ((event->type == GDK_BUTTON_PRESS) && (event->button == 1)) // left-button
	{
		if (goTimerActivate.elapsed() > 0
			&& goTimerActivate.elapsed() < 0.5) // ignore event when a control has been clicked in the last 0.5 sec
		{
			return false;
		}
		else
			return Gtk::TreeView::on_button_press_event(event);
	}
	
	bool bRetValue = Gtk::TreeView::on_button_press_event(event);
		
	// on double-click trigger event
	if (event->type == GDK_2BUTTON_PRESS 
		&& event->button == 1 // left button
		&& !(event->state & GDK_MODIFIER_MASK))
	{
		if (!grefSortedModel)
			return bRetValue;
		
		if (goTimerActivate.elapsed() > 0
			&& goTimerActivate.elapsed() < 0.5) // ignore event when a control has been double-clicked in the last 0.5 sec
		{
			return bRetValue;
		}
		
		// get selection and trigger dclick event
		VolumeGuiState* poSelVolume = getSelectedVolume();
		
		if (poSelVolume)
			m_signal_dclick(poSelVolume);
	}
	else if ((event->type == GDK_BUTTON_PRESS) && (event->button == 3)) // popup-menu
	{
		// get selection and trigger popup menu event
		VolumeGuiState* poSelVolume = getSelectedVolume();
		
		if (poSelVolume)
			m_signal_popup_menu(poSelVolume, event);
	}
	
	return bRetValue;
}

//==============================================================================

void VolumeTreeView::on_cell_activate(  const int iIndex,
										const int ix,
										const int iy)
{  
	if (!gpoVolumeCellRenderer || !grefSortedModel)
		return;

	//
	// emit buttons signal
	//
	
	VolumeCellRenderer::tLayoutCtrl tLayoutCtrlItem;

	// get visible row
	Gtk::TreeModel::Path start_path;
	Gtk::TreeModel::Path end_path;

	if (!get_visible_range(start_path, end_path))
		return;

	/*
	 * the following code doesn't compile on gtkmm 2.12.7
	if (!start_path ||  !end_path)
		return;*/
		
	const Gtk::TreeModel::iterator ite_start_path = grefSortedModel->get_iter(start_path);
	const Gtk::TreeModel::iterator ite_end_path = grefSortedModel->get_iter(end_path);

	if (ite_start_path && ite_end_path)
	{
		int iStartIndex = ((Gtk::TreeModel::Row)*ite_start_path)[goColumns.goColIndex];
		int iEndIndex = ((Gtk::TreeModel::Row)*ite_end_path)[goColumns.goColIndex];
		
		// reduce number of redraw
		if (gpoVolumeCellRenderer->isControl(   ix, 
												iy,
												iStartIndex,
												iEndIndex,
												&tLayoutCtrlItem))
		{
			queue_draw();
		}
			
		// check if cursos is over a control
		if (gpoVolumeCellRenderer->isControl(   ix, 
												iy,
												iStartIndex,
												iEndIndex,
												&tLayoutCtrlItem))
		{
			Glib::ustring stToolTip;
		
			goTimerActivate.reset();
			goTimerActivate.start();
		
			if (tLayoutCtrlItem.eBtnStateItem != VolumeCellRenderer::kBtnDisabled)
			{
				switch (tLayoutCtrlItem.eLayoutCtrlTypeItem)
				{
					case VolumeCellRenderer::kBtnCancel:
						debug_message(DEBUG_INFO, "Index: %d, emit cancel event", tLayoutCtrlItem.iIndex);
						
						m_signal_cancel(getVolumeByIndex(tLayoutCtrlItem.iIndex));
						break;
				
					case VolumeCellRenderer::kBtnEject:
						debug_message(DEBUG_INFO, "Index: %d, emit eject event", tLayoutCtrlItem.iIndex);
						
						m_signal_eject(getVolumeByIndex(tLayoutCtrlItem.iIndex));
						break;
				
					case VolumeCellRenderer::kBtnMount:
						debug_message(DEBUG_INFO, "Index: %d, emit mount event", tLayoutCtrlItem.iIndex);
						
						m_signal_mount(getVolumeByIndex(tLayoutCtrlItem.iIndex));
						break;
				
					default:
						break;
				}
			}
		}
	}
}

//==============================================================================

bool VolumeTreeView::on_key_press_event(GdkEventKey* event)
{
	if (!event)
		return Gtk::TreeView::on_key_press_event(event);
		
	if (event->keyval == GDK_Return)
		m_signal_key_enter_pressed(getSelectedVolume());
	
	return Gtk::TreeView::on_key_press_event(event);
}

//==============================================================================

bool VolumeTreeView::on_motion_notify_event(GdkEventMotion* event)
{	
	if (!gpoVolumeCellRenderer || !grefSortedModel)
		return Gtk::TreeView::on_motion_notify_event(event);

	//
	// show tooltip
	// 
	if (QConfig::getInstance().isHideToolTips())
		return Gtk::TreeView::on_motion_notify_event(event);
	
	set_has_tooltip(false);

	VolumeCellRenderer::tLayoutCtrl tLayoutCtrlItem;

	// get visible row
	Gtk::TreeModel::Path start_path;
	Gtk::TreeModel::Path end_path;

	if (!get_visible_range(start_path, end_path))
		return Gtk::TreeView::on_motion_notify_event(event);	

	/*
	 * the following conde doesn't compile on gtkmm 2.12.7
	if (!start_path ||  !end_path)
	if (!start_path ||  !end_path)
		return Gtk::TreeView::on_motion_notify_event(event);*/
		
	const Gtk::TreeModel::iterator ite_start_path = grefSortedModel->get_iter(start_path);
	const Gtk::TreeModel::iterator ite_end_path = grefSortedModel->get_iter(end_path);

	if (ite_start_path && ite_end_path)
	{
		int iStartIndex = ((Gtk::TreeModel::Row)*ite_start_path)[goColumns.goColIndex];
		int iEndIndex = ((Gtk::TreeModel::Row)*ite_end_path)[goColumns.goColIndex];
		
		// reduce number of redraw
		if (gpoVolumeCellRenderer->isControl(   event->x, 
												event->y,
												iStartIndex,
												iEndIndex,
												&tLayoutCtrlItem))
		{
			queue_draw();
		}
					
		// check if cursos is over a control
		if (gpoVolumeCellRenderer->isControl(   event->x, 
												event->y,
												iStartIndex,
												iEndIndex,
												&tLayoutCtrlItem))
		{
			if (tLayoutCtrlItem.eBtnStateItem != VolumeCellRenderer::kBtnDisabled)
			{
				Glib::ustring stToolTip;
		
				switch (tLayoutCtrlItem.eLayoutCtrlTypeItem)
				{
					case VolumeCellRenderer::kBtnCancel:
						stToolTip = _("Abort");
						break;
					case VolumeCellRenderer::kBtnEject:
						stToolTip = _("Eject");
						break;
					case VolumeCellRenderer::kBtnMount:
						stToolTip = _("Mount");
						break;
					default:
						break;
				}
		
				if (!stToolTip.empty())
				{
					set_has_tooltip(true);
					set_tooltip_text(stToolTip);
				}
			}
		}
	}
	
	return Gtk::TreeView::on_motion_notify_event(event);
}

//==============================================================================

void VolumeTreeView::on_rows_reordered( const Gtk::TreeModel::Path& path, 
										const Gtk::TreeModel::iterator& iter, 
										int* new_order)
{
	sortVolumeCellRenderer();
}

//==============================================================================

void VolumeTreeView::on_selection_changed()
{
	m_signal_selection_changed(getSelectedVolume());
}

//==============================================================================

bool VolumeTreeView::on_timeout()
{
	goTimerActivate.stop();
	goTimerActivate.reset();	
	
	return true;
}

//==============================================================================

void VolumeTreeView::removeVolume(VolumeGuiState *poVolumeGuiState)
{
	Gtk::TreeModel::Row oRow;
	std::vector<int> vecIndexMatch;
	
	// get all match index
	for (   std::list<tVolumeGuiState>::iterator ite = glstVolumeGuiState.begin();
			ite != glstVolumeGuiState.end();
			++ite)
	{
		if (((tVolumeGuiState)*ite).poVolumeGuiState == poVolumeGuiState)
			vecIndexMatch.push_back(((tVolumeGuiState)*ite).iIndex);
	}
	
	// remove row(s)
	for (   std::vector<int>::size_type xi = 0;
			xi < vecIndexMatch.size();
			++xi)
	{
		if (getRowByIndex(vecIndexMatch[xi], &oRow))
		{
			if (gpoVolumeCellRenderer)
				gpoVolumeCellRenderer->removeRowIndex(vecIndexMatch[xi]);
				
			grefTreeModel->erase(oRow);
		}
	}
}

//==============================================================================

void VolumeTreeView::sortByLabel()
{
	if (!grefSortedModel)
		return;
	
	grefSortedModel->set_sort_column(goColumns.goColDrawLabel, Gtk::SORT_ASCENDING);
	set_model(grefSortedModel);
}

//==============================================================================

void VolumeTreeView::sortByPos()
{
	if (!grefSortedModel)
		return;
	
	grefSortedModel->set_sort_column(goColumns.goColPos, Gtk::SORT_ASCENDING);
	set_model(grefSortedModel);
}

//==============================================================================

void VolumeTreeView::sortVolumeCellRenderer()
{
	if (!gpoVolumeCellRenderer || !grefSortedModel)
		return;
		
	// sort volume cell rendered index
	typedef Gtk::TreeModel::Children type_children; //minimise code length.

	type_children children = grefSortedModel->children();
	
	int ci = 0;
	
	for (   type_children::iterator ite = children.begin();
			ite != children.end(); 
			++ite)
	{
		Gtk::TreeModel::Row row_ite = *ite;
		
		gpoVolumeCellRenderer->setIndexOrder(row_ite[goColumns.goColIndex], ci);
			
		ci++;
	}
}

//==============================================================================

void VolumeTreeView::updateVolume(VolumeGuiState *poVolumeGuiState)
{
	Gtk::TreeModel::Row oRow;
	std::vector<int> vecIndexMatch;
	
	// get all match index
	for (   std::list<tVolumeGuiState>::iterator ite = glstVolumeGuiState.begin();
			ite != glstVolumeGuiState.end();
			++ite)
	{
		if (((tVolumeGuiState)*ite).poVolumeGuiState == poVolumeGuiState)
			vecIndexMatch.push_back(((tVolumeGuiState)*ite).iIndex);
	}
	
	// update row
	for (   std::vector<int>::size_type xi = 0;
			xi < vecIndexMatch.size();
			++xi)
	{		
		getRowByIndex(vecIndexMatch[xi], &oRow);
		
		mapVolumeGuiState(oRow, poVolumeGuiState);
	}
	
	sortVolumeCellRenderer();
}

//==============================================================================
